// Copyright 2007 Hiroyuki Ogasawara
// vim:ts=4 sw=4:
/*!
	@file
	@author	Hiroyuki Ogasawara
	@date	2007/10/25
	@brief
		`̃sNZ\[gVF[_[

	@sa http://dench.flatlib.jp/
	@sa http://wlog.flatlib.jp/
*/



Texture2D		texDiffuse;
SamplerState	smpDiffuse;

cbuffer cb_GCamera {
	float4x4	ViewProjection;
	float3		CameraPosition;
};

cbuffer cb_GCameraFrame
{
	float4x4	World;
	float		cFrame;
	uint		Sorting;
};


DepthStencilState Zdisable
{
	DepthEnable= FALSE;
};

RasterizerState CullNone
{
	CullMode= NONE;
};

RasterizerState CullBack
{
	CullMode= BACK;
};


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

#define		MAT_SPECULARCOLOR	float3( 1.0f, 1.0f, 1.0f )
#define		COL_BGCOLOR			float3( 0.0f, 0.0f, 0.3f )

#define		MAXTORUS			16
#define		__PI				3.1415926535f

#define		INPUTRESOURCES		4


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------


struct VS_INPUT {
	float3	Pos			: POSITION;
	float3	Normal		: NORMAL;
	float2	Tex0		: TEXCOORD0;
};


struct VS_OUTPUT {
	float4	Pos			: SV_POSITION;
	float3	Normal		: NORMAL;
	float2	Tex0		: TEXCOORD0;
	float3	WorldPos	: WORLDPOS;
	float3	Color		: COLOR;
};


typedef	VS_OUTPUT	GS_OUTPUT;
typedef	GS_OUTPUT	PS_INPUT;


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

static const float3	_Colors[]= {
	{	1.0f, 1.0f, 0.0f,	},
	{	1.0f, 0.0f, 0.0f,	},
	{	0.0f, 0.5f, 1.0f,	},
	{	1.0f, 0.0f, 1.0f,	},
	{	0.0f, 1.0f, 0.0f,	},
	{	0.0f, 1.0f, 1.0f,	},
	{	0.5f, 0.5f, 0.5f,	},
	{	1.0f, 0.5f, 0.5f,	},
};

/*!
	@brief
		f`
*/
VS_OUTPUT VS_Main( VS_INPUT In, uint insid : SV_InstanceID )
{
	// scaling
	In.Pos.xyz *= 0.4f;

	// animation
	VS_OUTPUT	Out;
	float4x4	wmat= World;
	float4		sc;
	sincos( cFrame * 0.005f + insid*(2*__PI/MAXTORUS) , sc.x, sc.y );
	sincos( insid*(2*__PI/MAXTORUS)*4, sc.z, sc.w );
	float4	wpos= mul( float4( In.Pos.xyz, 1.0f ), wmat );
	wpos.xyz+= sc.yzx * float3( 10.0f, 8.0f, 10.0f );

	// transform
	Out.WorldPos= wpos;
	Out.Pos= mul( wpos, ViewProjection );
	Out.Normal= mul( In.Normal.xyz, (float3x3)World );
	Out.Tex0= In.Tex0;
	Out.Color= _Colors[ insid ];

	return	Out;
}


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

/*!
	@brief
		Rendering Sample
*/
void PS_Main( PS_INPUT In,
				out float4 Ocolor, out float Odepth )
{
	const float	_BlendAlpha= 0.50f;	// =Texture.Sample().w;

	float3	normal= normalize( In.Normal.xyz );
	float3	eye= normalize( CameraPosition.xyz - In.WorldPos.xyz );
	float3	lvect= normalize( CameraPosition.xyz - In.WorldPos.xyz );
	float	sdot= saturate( dot( normal, normalize( eye.xyz + lvect.xyz ) ) );
	float	diffuse= saturate( dot( normal, lvect ) );
	float	specular= pow( sdot, 80.0f );
	float3	color= In.Color * diffuse + MAT_SPECULARCOLOR * specular;

	Ocolor= saturate( float4( color.xyz, _BlendAlpha ) ) * 0.99f;
	Odepth= In.Pos.z;
}


/*!
	@brief
		Encode 3 layer
*/
void PS_Color3( PS_INPUT In,
		out float4 o0 : SV_Target0,
		out float4 o1 : SV_Target1 )
{
	float4	color;
	float	depth;
	PS_Main( In, color, depth );

	// encode
	o0= float4( (uint3)( saturate( color.xyz ) * 255 ).xyz, 256.0f );

	uint	iz= (uint)( depth * 65535 );
	uint3	icolor2= uint3( (iz >> 8) & 255, iz & 255, color.w * 255 );
	o1= float4( icolor2.xyz, 256.0f );
}


/*!
	@brief
		Encode 6 layer
*/
void PS_Color6( PS_INPUT In,
		bool isfront : SV_IsFrontFace,
		out float4 o0 : SV_Target0,
		out float4 o1 : SV_Target1,
		out float4 o2 : SV_Target2,
		out float4 o3 : SV_Target3 )
{
	if( !isfront ){
		In.Normal= -In.Normal;
	}

	float4	color;
	float	depth;
	PS_Main( In, color, depth );

	// encode
	const int	EncodeBit= 128;
	const int	ColorMask= (EncodeBit-1);
	const int	DepthMask= (EncodeBit*EncodeBit-1);
	o0= float4( (uint3)( saturate( color.xyz ) * ColorMask ).xyz, 256.0f );

	uint	iz= (uint)( In.Pos.z * DepthMask );
	uint3	icolor2= uint3( (iz >> 7) & ColorMask,
									iz & ColorMask, color.w * ColorMask );
	o1= float4( icolor2.xyz, 256.0f );
	o2= float4( o0.xyz, 1.0f/256.0f );
	o3= float4( o1.xyz, 1.0f/256.0f );
}


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

Texture2D		InputTexture[INPUTRESOURCES];
SamplerState	inputstate
{
	Filter= MIN_MAG_MIP_POINT;
	AddressU= Clamp;
	AddressV= Clamp;
};

static const float4	vertex[4]= {
	{	-1.0f,	 1.0f,	0.0f,	1.0f,	},
	{	 1.0f,	 1.0f,	0.0f,	1.0f,	},
	{	-1.0f,	-1.0f,	0.0f,	1.0f,	},
	{	 1.0f,	-1.0f,	0.0f,	1.0f,	},
};
static const float2	uv[4]= {
	{	0.0f,	0.0f,	},
	{	1.0f,	0.0f,	},
	{	0.0f,	1.0f,	},
	{	1.0f,	1.0f,	},
};


/*!
	@brief
		generate vertex
*/
void VS_Sort( uint vid : SV_VertexID,
			out float4 opos : SV_POSITION, out float2 ouv : TEXCOORD )
{
	opos= vertex[ vid ];
	ouv= uv[ vid ];
}


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

#define	_SORT3_MAX_LAYER	3


/*!
	@brief
		sorting 3 layer
*/
float4 PS_Sort3( float4 Pos : SV_POSITION, float2 UV : TEXCOORD ) : SV_Target
{
	float4	color= InputTexture[0].Sample( inputstate, UV );
	float4	color2= InputTexture[1].Sample( inputstate, UV );
	float4	b[_SORT3_MAX_LAYER];
	float	z[_SORT3_MAX_LAYER];

	// decode
	uint	id= 0;
	for(; color.w > 1.0f && id< _SORT3_MAX_LAYER ; id++ ){
		color.w*= 1.0f / 256.0f;
		float		wscale= 1.0f / color.w;
		const float	bscale= 1.0f / 255.0f;
		b[id].xyz= ((uint3)( color * wscale ) & 255) * bscale;
		uint3	ui= ((uint3)( color2 * wscale ) & 255);
		b[id].w= ui.z * bscale;		// alpha
		z[id]= ui.x*256 + ui.y;	// z, scalingȂ
	}

	float3	bcolor= COL_BGCOLOR;	// bgcolor

	if( Sorting ){

		// sort
		float4	bo[_SORT3_MAX_LAYER];
		for( uint i= 0 ; i< id ; i++ ){
			float	bz= 0.0f;
			uint	bi= 0;
			for( uint j= 0 ; j< id ; j++ ){
				if( z[j] > bz ){
					bz= z[j];
					bi= j;
				}
			}
			bo[i]= b[bi];
			z[bi]= 0.0f;
		}

		for( uint i= 0 ; i< id ; i++ ){
			bcolor= lerp( bcolor, bo[i], bo[i].w );
		}

	}else{
		for( uint i= 0 ; i< id ; i++ ){
			bcolor= lerp( bcolor, b[i], b[i].w );
		}
	}

	return	float4( bcolor.xyz, 1.0f );
}


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

#define	_SORT6_MAX_LAYER	6
#define	_SORT6_PP_LAYER		3

/*!
	@brief
		sorting 6 layer
*/
float4 PS_Sort6( float4 Pos : SV_POSITION, float2 UV : TEXCOORD ) : SV_Target
{
	float4	color= InputTexture[0].Sample( inputstate, UV );
	float4	color2= InputTexture[1].Sample( inputstate, UV );
	float4	color3= InputTexture[2].Sample( inputstate, UV );
	float4	color4= InputTexture[3].Sample( inputstate, UV );
	float4	b[_SORT6_MAX_LAYER];
	float	z[_SORT6_MAX_LAYER];

	// decode
	uint	id= 0;
	for(; color.w > 1.0f && id< _SORT6_PP_LAYER ; id++ ){
		color.w*= 1.0f / 256.0f;
		float		wscale= 1.0f / color.w;
		const float	bscale= 2.0f / 255.0f;
		b[id].xyz= ((uint3)( color * wscale ) & 127) * bscale;
		uint3	ui= ((uint3)( color2 * wscale ) & 127);
		b[id].w= ui.z * bscale;		// alpha
		z[id]= ui.x*128 + ui.y;		// z, scalingȂ
	}

	float	dep= -log2( color3.w ) * (1.0f/8.0f) + 0.5f;
	uint	id2= (uint)clamp( dep-3.0f, 0.0f, 3.0f );

	float	ae= 1.0f;
	for( uint i= 0 ; i< id2 ; i++, id++ ){
		const float	bscale= 2.0f / 255.0f;
		float	wscale= ae;
		b[id].xyz= ((uint3)( color3 * wscale ) & 127) * bscale;
		uint3	ui= ((uint3)( color4 * wscale ) & 127);
		b[id].w= ui.z * bscale;		// alpha
		z[id]= ui.x*128 + ui.y;		// z, scalingȂ
		ae*= 256.0f;
	}

	float3	bcolor= COL_BGCOLOR;	// bgcolor

	if( Sorting ){

		// sort
		float4	bo[_SORT6_MAX_LAYER];
		for( uint i= 0 ; i< id ; i++ ){
			float	bz= 0.0f;
			uint	bi= 0;
			for( uint j= 0 ; j< id ; j++ ){
				if( z[j] > bz ){
					bz= z[j];
					bi= j;
				}
			}
			bo[i]= b[bi];
			z[bi]= 0.0f;
		}

		for( uint i= 0 ; i< id ; i++ ){
			bcolor= lerp( bcolor, bo[i], bo[i].w );
		}

	}else{
		for( uint i= 0 ; i< id ; i++ ){
			bcolor= lerp( bcolor, b[i], b[i].w );
		}
	}

	return	float4( bcolor.xyz, 1.0f );
}



//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

VertexShader	vs_main= CompileShader( vs_4_0, VS_Main() );
PixelShader		ps_color3= CompileShader( ps_4_0, PS_Color3() );
PixelShader		ps_color6= CompileShader( ps_4_0, PS_Color6() );


technique10 Main3
{
	pass P0
	{
		SetDepthStencilState( Zdisable, 0 );
//		SetRasterizerState( CullNone );
		SetVertexShader( vs_main );
		SetGeometryShader( NULL );
		SetPixelShader( ps_color3 );
	}
}


technique10 Main6
{
	pass P0
	{
		SetDepthStencilState( Zdisable, 0 );
//		SetRasterizerState( CullNone );
		SetVertexShader( vs_main );
		SetGeometryShader( NULL );
		SetPixelShader( ps_color6 );
	}
}


VertexShader	vs_sort= CompileShader( vs_4_0, VS_Sort() );
PixelShader		ps_sort3= CompileShader( ps_4_0, PS_Sort3() );
PixelShader		ps_sort6= CompileShader( ps_4_0, PS_Sort6() );

technique10 Sort3
{
	pass P0
	{
		SetDepthStencilState( Zdisable, 0 );
		SetRasterizerState( CullBack );
		SetVertexShader( vs_sort );
		SetGeometryShader( NULL );
		SetPixelShader( ps_sort3 );
	}
}


technique10 Sort6
{
	pass P0
	{
		SetDepthStencilState( Zdisable, 0 );
		SetRasterizerState( CullBack );
		SetVertexShader( vs_sort );
		SetGeometryShader( NULL );
		SetPixelShader( ps_sort6 );
	}
}


